#include "stack.h"
#include "vec.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum { NOT_VISITED = 0, IN_STACK = 1, DONE = 2 };

graph_vec vnew_graph(void)
{
	graph_vec g;
	g.n = -1;
	g.n = -1;
	g.cap = 10;
	g.node = malloc(10 * sizeof(nodept));
	assert(g.node);
	return g;
}

void vinsert_node(graph_vec *g, fix_str str)
{
	++g->n;
	if (g->n == g->cap)
		g->node = reallocarray(g->node, (g->cap += 128), sizeof(nodept));
	struct node *n = malloc(sizeof(struct node));
	assert(n);
	n->num = g->n;
	n->nbs = 0;
	n->vu = 0;
	n->dvu = 0;
	n->cap = 5;
	n->succ = malloc(5 * sizeof(nodept));
	strcpy(n->func, str);
	g->node[g->n] = n;
}

void vdelete_graph(graph_vec *g)
{
	int i;
	for (i = 0; i <= g->n; ++i) {
		if (g->node[i]) {
			if (g->node[i]->succ) {
				free(g->node[i]->succ);
				g->node[i]->succ = NULL;
			}
			free(g->node[i]);
			g->node[i] = NULL;
		}
	}
	free(g->node);
}

int vis_in(graph_vec g, fix_str func)
{
	int i;
	if (g.n < 0) {
		return -1;
	}
	for (i = 0; i <= g.n; i++) {
		if (strcmp(g.node[i]->func, func) == 0) {
			return i;
		}
	}
	return -1;
}

static void vprint_succ(nodept node)
{
	printf("  Succ: %s\n", node->func);
}

void vprint_nodes(nodept node)
{
	int i;
	if (node->vu)
		return;
	node->vu = 1;
	printf("NODE: %s\n", node->func);
	for (i = 0; i < node->nbs; ++i) {
		vprint_succ(node->succ[i]);
	}
	for (i = 0; i < node->nbs; ++i) {
		if (node->succ[i]->nbs != 0)
			vprint_nodes(node->succ[i]);
	}
}

static void vwrite_node_dot(FILE *fp, graph_vec g)
{
	int i, j;
	for (i = 0; i <= g.n; i++) {
		for (j = 0; j < g.node[i]->nbs; j++) {
			// printf("\t%s -> %s;\n",  node->func, node->succ[i]->func);
			fprintf(fp, "\t%s -> %s;\n", g.node[i]->func, g.node[i]->succ[j]->func);
		}
	}
}

static void main_graph(FILE *fp, nodept node)
{
	int i;
	if (node->vu == 1)
		return;
	node->vu = 1;
	// printf("NODE %d: %s\n", node->num, node->func);
	for (i = 0; i < node->nbs; ++i) {
		fprintf(fp, "\t%s -> %s;\n", node->func, node->succ[i]->func);
	}
	for (i = 0; i < node->nbs; ++i) {
		main_graph(fp, node->succ[i]);
	}
}

void vdot_graph(graph_vec g, bool main)
{
	FILE *fp, *fp1;
	int i;

	fp = fopen("graph.dot", "w");
	fp1 = fopen("CC.dot", "w");

	fprintf(fp, "digraph CCmain {\n");
	fprintf(fp1, "digraph CC {\n");

	if (main) {
		fix_str m = "main";
		int n = vis_in(g, m);
		if (n != -1) {
			main_graph(fp, g.node[n]);
		} else {
			fprintf(stderr, "main not found.\n");
			exit(1);
		}
		for (i = 0; i <= g.n; ++i) {
			if (!g.node[i]->vu) {
				main_graph(fp1, g.node[i]);
			}
		}

	} else {
		vwrite_node_dot(fp, g);
	}

	fprintf(fp, "}");
	fprintf(fp1, "}");

	fclose(fp);
	fclose(fp1);
}

// NOTE: https://www.baeldung.com/cs/detecting-cycles-in-directed-graph
static void print_cycle(graph_vec g, stack_t s, int v)
{
	stack_t s2 = init_stack();
	push(&s2, pop(&s));
	while (top(s2) != v) {
		push(&s2, pop(&s));
	}
	while (!empty(s2)) {
		printf("%s -> ", g.node[pop(&s2)]->func);
	}
	printf("%s\n", g.node[v]->func);
	free_stack(&s2);
}

static void processDFS(nodept n, stack_t *s, graph_vec g)
{
	int i;
	for (i = 0; i < n->nbs; ++i) {
		int v = n->succ[i]->num;
		if (n->succ[i]->dvu == IN_STACK) {
			print_cycle(g, *s, v);
		} else if (n->succ[i]->dvu == NOT_VISITED) {
			push(s, v);
			n->succ[i]->dvu = IN_STACK;
			processDFS(n->succ[i], s, g);
		}
	}
	n->dvu = DONE;
	pop(s);
}

void vfind_cycles(graph_vec *g)
{
	int v;
	for (v = 0; v <= g->n; ++v) {
		if (g->node[v]->dvu == NOT_VISITED) {
			stack_t s = init_stack();
			push(&s, v);
			g->node[v]->dvu = IN_STACK;
			processDFS(g->node[v], &s, *g);
			free_stack(&s);
		}
	}
}
